from hotsos.core.plugins.openstack.common import OpenstackChecksBase
from hotsos.core.plugins.openstack.neutron import NeutronHAInfo


class OpenstackSummary(OpenstackChecksBase):
    summary_part_index = 0

    def __0_summary_release(self):
        return {'name': self.release_name,
                'days-to-eol': self.days_to_eol}

    def __1_summary_services(self):
        """Get string info for running services."""
        if self.systemd.services:
            return self.systemd.summary
        if self.pebble.services:
            return self.pebble.summary

    def __2_summary_dpkg(self):
        # require at least one core package to be installed to include
        # this in the report.
        if self.apt.core:
            return self.apt.all_formatted

    def __3_summary_docker_images(self):
        # require at least one core image to be in-use to include
        # this in the report.
        if self.docker.core:
            return self.docker.all_formatted

    def __4_summary_neutron_l3ha(self):
        routers = {}
        ha_info = NeutronHAInfo()
        for router in ha_info.ha_routers:
            state = router.ha_state
            if state in routers:
                routers[state] += 1
            else:
                routers[state] = 1

        if routers:
            return routers



from hotsos.core.config import HotSOSConfig
from hotsos.core.issues import IssuesManager, HotSOSScenariosWarning
from hotsos.core.log import log
from hotsos.core.search import (
    FileSearcher,
    SearchConstraintSearchSince,
)
from hotsos.core.ycheck.engine import (
    YDefsLoader,
    YDefsSection,
    YHandlerBase,
)
from hotsos.core.ycheck.engine.properties.search import (
    CommonTimestampMatcher,
)


class Scenario(object):
    def __init__(self, name, checks, conclusions):
        log.debug("scenario: %s", name)
        self.name = name
        self._checks = checks
        self._conclusions = conclusions

    @property
    def checks(self):
        return {c.name: c for c in self._checks}

    @property
    def conclusions(self):
        return {c.name: c for c in self._conclusions}


class YScenarioChecker(YHandlerBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._scenarios = []
        hours = 24
        if HotSOSConfig.use_all_logs:
            hours *= HotSOSConfig.max_logrotate_depth

        constraint = SearchConstraintSearchSince(
                                        ts_matcher_cls=CommonTimestampMatcher,
                                        hours=hours)
        self._searcher = FileSearcher(constraint=constraint)

    @property
    def searcher(self):
        return self._searcher

    def load(self):
        plugin_content = YDefsLoader('scenarios').plugin_defs
        if not plugin_content:
            return

        yscenarios = YDefsSection(HotSOSConfig.plugin_name, plugin_content)
        if (not HotSOSConfig.force_mode and yscenarios.requires and not
                yscenarios.requires.result):
            log.debug("plugin '%s' scenarios pre-requisites not met - "
                      "skipping", HotSOSConfig.plugin_name)
            return

        log.debug("sections=%s, scenarios=%s",
                  len(list(yscenarios.branch_sections)),
                  len(list(yscenarios.leaf_sections)))

        to_skip = set()

        checks = []
        for scenario in yscenarios.leaf_sections:
            fullname = "{}.{}.{}".format(HotSOSConfig.plugin_name,
                                         scenario.parent.name, scenario.name)
            if (HotSOSConfig.scenario_filter and
                    fullname != HotSOSConfig.scenario_filter):
                log.info("skipping scenario %s (filter=%s)", fullname,
                         HotSOSConfig.scenario_filter)
                continue

            # Only register scenarios if requirements are satisfied.
            group_name = scenario.parent.name
            if (not HotSOSConfig.force_mode and
                    (group_name in to_skip or
                        (scenario.requires and not scenario.requires.result))):
                log.debug("%s requirements not met - skipping scenario %s",
                          group_name, scenario.name)
                to_skip.add(group_name)
                continue

            checks.append(scenario.checks)
            scenario.checks.initialise(scenario.vars, scenario.input,
                                       self.searcher, scenario)
            _scenario = Scenario(scenario.name, scenario.checks,
                                 scenario.conclusions)
            scenario.conclusions.initialise(scenario.vars, _scenario.checks)
            self._scenarios.append(_scenario)

        log.debug("executing check searches")
        results = self.searcher.run()
        for check in checks:
            check.check_context.search_results = results

    @property
    def scenarios(self):
        return self._scenarios

    def _run_scenario_conclusion(self, scenario, issue_mgr):
        """ Determine the conclusion of this scenario. """
        results = {}
        # Run conclusions in order of priority. One or more conclusion may
        # share the same priority. If one or more conclusion of the same
        # priority is reached the rest (of lower priority) are ignored.
        last_priority = None
        for conc in sorted(scenario.conclusions.values(), key=lambda _conc:
                           int(_conc.priority or 1), reverse=True):
            priority = int(conc.priority or 1)
            if last_priority is not None:
                if priority < last_priority and last_priority in results:
                    break

            last_priority = priority
            if conc.reached(scenario.checks):
                if priority in results:
                    results[priority].append(conc)
                else:
                    results[priority] = [conc]

                log.debug("conclusion reached: %s (priority=%s)", conc.name,
                          priority)

        if results:
            highest = max(results.keys())
            log.debug("selecting highest priority=%s conclusions (%s)",
                      highest, len(results[highest]))
            for conc in results[highest]:
                issue_mgr.add(conc.issue, context=conc.issue_context)
        else:
            log.debug("no conclusions reached")

    def run(self):
        failed_scenarios = []
        issue_mgr = IssuesManager()
        for scenario in self.scenarios:
            log.debug("running scenario: %s", scenario.name)
            # catch failed scenarios and allow others to run
            try:
                self._run_scenario_conclusion(scenario, issue_mgr)
            except Exception:
                log.exception("caught exception when running scenario %s:",
                              scenario.name)
                failed_scenarios.append(scenario.name)

        if failed_scenarios:
            msg = ("One or more scenarios failed to run ({}) - run hotsos in "
                   "debug mode (--debug) to get more detail".
                   format(', '.join(failed_scenarios)))
            issue_mgr.add(HotSOSScenariosWarning(msg))

    def load_and_run(self):
        self.load()
        return self.run()



from hotsos.core.plugins.openstack.common import (
    OpenstackEventCallbackBase,
    OpenstackEventHandlerBase,
)
from hotsos.core.search import (
    FileSearcher,
    SearchDef,
    SearchConstraintSearchSince,
)
from hotsos.core.ycheck.engine.properties.search import CommonTimestampMatcher

EXT_EVENT_META = {'network-vif-plugged': {'stages_keys':
                                          ['Preparing', 'Received',
                                           'Processing']},
                  'network-changed': {'stages_keys': ['Received',
                                                      'Refreshing']}}


class ExternalEventsCallback(OpenstackEventCallbackBase):
    event_group = 'nova.external-events'
    event_names = ['network-changed', 'network-vif-plugged']

    def _get_state_dict(self, event_name):
        state = {}
        for key in EXT_EVENT_META[event_name]['stages_keys']:
            state[key] = False

        return state

    def __call__(self, event):
        ext_output = {}
        events = {}
        events_found = {}

        c = SearchConstraintSearchSince(ts_matcher_cls=CommonTimestampMatcher)
        s = FileSearcher(constraint=c)
        for result in event.results:
            instance_id = result.get(1)
            event_id = result.get(3)
            result_path = event.searcher.resolve_source_id(result.source_id)
            events[event_id] = {'instance_id': instance_id,
                                'data_source': result_path}

            for stage in EXT_EVENT_META[event.name]['stages_keys']:
                expr = (r".+\[instance: {}\]\s+{}\s.*\s?event\s+{}-{}.? "
                        ".+".
                        format(instance_id, stage, event.name, event_id))
                tag = "{}_{}_{}".format(instance_id, event_id, stage)
                sd = SearchDef(expr, tag, hint=event.name,
                               store_result_contents=False)
                s.add(sd, result_path)

        results = s.run()
        for event_id, event_dict in events.items():
            instance_id = event_dict['instance_id']
            data_source = event_dict['data_source']
            stages = self._get_state_dict(event.name)
            for stage in stages:
                tag = "{}_{}_{}".format(instance_id, event_id, stage)
                r = results.find_by_tag(tag, path=data_source)
                if r:
                    stages[stage] = True

            if all(values for stage, values in stages.items()):
                result = 'succeeded'
            else:
                result = 'failed'

            if result not in ext_output:
                ext_output[result] = []

            info = {'port': event_id, 'instance': instance_id}
            ext_output[result].append(info)

        if ext_output:
            for result, values in ext_output.items():
                events_found[result] = list(values)

        return events_found


class NovaExternalEventChecks(OpenstackEventHandlerBase):
    event_group = 'nova.external-events'
    summary_part_index = 1

    def __8_summary_os_server_external_events(self):
        return self.load_and_run()



from propertree.propertree2 import PTreeLogicalGrouping
from hotsos.core.log import log
from hotsos.core.ycheck.engine.properties.common import (
    YPropertyMappedOverrideBase,
)
from hotsos.core.ycheck.engine.properties.requires.types import (
    apt,
    snap,
    config,
    pebble,
    systemd,
    property as rproperty,
    path,
    varops,
)

CACHE_CHECK_KEY = '__PREVIOUSLY_CACHED_PROPERTY_TYPE'


def cache_result(cache, item, result, grouped=False):
    """
    Also copy the cache of the item into the local cache. If part of a
    logical group whereby more than one item is being evaluated to get
    a final results, we only copy the cache data if the result is True
    and  either (a) it is the first to be saved or (b) it has the same type
    as the previous one to be saved.

    For example with the following group:

    or:
      - apt: foo
      - apt: bar

    If package foo exists and bar does not, the final contents of the group
    will be that from 'apt: foo'. If both exist then it will be the result of
    'apt: bar'.

    @param item: YRequirementType object
    @param grouped: If True, item is part of a (logical) group of one or
                    more items or a list with more than one item to which
                    to default logical operator (AND) will be applied.
    @return: boolean result i.e. True or False
    """
    previously_cached = getattr(cache, CACHE_CHECK_KEY)
    req_prop_type = item.__class__.__name__
    if not grouped or (result and
                       previously_cached in (None, req_prop_type)):
        if previously_cached is None:
            cache.set(CACHE_CHECK_KEY, req_prop_type)

        log.debug("merging item (type=%s) cache id=%s into cache id=%s",
                  req_prop_type, item.cache.id, cache.id)
        cache.merge(item.cache)


class RequiresLogicalGrouping(PTreeLogicalGrouping):
    _override_autoregister = False

    def fetch_item_result(self, item):
        if isinstance(item, RequiresLogicalGrouping):
            item.cache = self._override_parent.cache

        result = super().fetch_item_result(item)
        if isinstance(item, RequiresLogicalGrouping):
            cache_result(self._override_parent.cache, item, result, True)

        return result


class YPropertyRequires(YPropertyMappedOverrideBase):
    _override_keys = ['requires']
    _override_members = [apt.YRequirementTypeAPT,
                         snap.YRequirementTypeSnap,
                         config.YRequirementTypeConfig,
                         pebble.YRequirementTypePebble,
                         systemd.YRequirementTypeSystemd,
                         rproperty.YRequirementTypeProperty,
                         path.YRequirementTypePath,
                         varops.YPropertyVarOps]
    # We want to be able to use this property both on its own and as a member
    # of other mapping properties e.g. Checks. The following setting enables
    # this.
    _override_auto_implicit_member = False
    _override_logical_grouping_type = RequiresLogicalGrouping

    @property
    def result(self):
        """
        Content can either be a single requirement, dict of requirement groups
        or list of requirements. List may contain individual requirements or
        groups.
        """
        log.debug("running requirement")
        try:
            results = []
            stop_executon = False
            for member in self.members:
                for item in member:
                    result = item.result
                    if not isinstance(item, RequiresLogicalGrouping):
                        cache_result(self.cache, item, result, grouped=False)

                    results.append(result)
                    if RequiresLogicalGrouping.is_exit_condition_met('and',
                                                                     result):
                        stop_executon = True
                        break

                if stop_executon:
                    break

            log.debug("%s results=%s", self.__class__.__name__, results)
            result = all(results)
            self.cache.set('passes', result)
            return result
        except Exception as exc:
            log.error("caught exception when running requirement: %s", exc)
            raise



import abc
import os
import re

from searchkit.utils import MPCache
from hotsos.core.config import HotSOSConfig
from hotsos.core.log import log


class NullCache(object):
    """ A cache that does nothing but maintains the MPCache abi. """

    def get(self, *args, **kwargs):
        log.debug("null cache get() op args=%s kwargs=%s", args, kwargs)

    def set(self, *args, **kwargs):
        log.debug("null cache set() op args=%s kwargs=%s", args, kwargs)


class HostHelpersBase(abc.ABC):

    def __init__(self, *args, **kwargs):
        if not self.cache_root or not os.path.exists(self.cache_root):
            log.debug("cache root invalid or does not exist so disabling %s "
                      "cache", self.__class__.__name__)
            self.cache = NullCache()
        else:
            self.cache = MPCache(self.cache_name,
                                 'host_helpers_{}'.format(self.cache_type),
                                 self.cache_root)

        super().__init__(*args, **kwargs)

    @property
    def cache_root(self):
        """
        By default caches are global to all plugins but this can be overridden
        if we want otherwise.
        """
        return HotSOSConfig.global_tmp_dir

    @property
    @abc.abstractmethod
    def cache_name(self):
        """ Unique name for cache used by instance of this object. """

    @property
    @abc.abstractmethod
    def cache_type(self):
        """
        Unique name for the type of cache used by instance of this object.
        """

    @abc.abstractmethod
    def cache_load(self):
        """ Load cache contents. """

    @abc.abstractmethod
    def cache_save(self):
        """ Save contents to cache. """


class ServiceManagerBase(abc.ABC):
    PS_CMD_EXPR_TEMPLATES = {
        'absolute': r".+\S+bin/({})(?:\s+.+|$)",
        'snap': r".+\S+\d+/({})(?:\s+.+|$)",
        'relative': r".+\s({})(?:\s+.+|$)",
    }

    def __init__(self, service_exprs, ps_allow_relative=True):
        """
        @param service_exprs: list of python.re expressions used to match
                              service names.
        @param ps_allow_relative: whether to allow commands to be identified
                                  from ps as run using an relative binary
                                  path e.g. mycmd as opposed to /bin/mycmd.
        """
        self._ps_allow_relative = ps_allow_relative
        self._service_exprs = set(service_exprs)

    def get_cmd_from_ps_line(self, line, expr):
        """
        Match a command in ps output line.

        @param line: line from ps output
        @param expr: regex to match a command. See PS_CMD_EXPR_TEMPLATES.
        @param return: matched command name.
        """
        for expr_type, expr_tmplt in self.PS_CMD_EXPR_TEMPLATES.items():
            if expr_type == 'relative' and not self._ps_allow_relative:
                continue

            ret = re.compile(expr_tmplt.format(expr)).match(line)
            if ret:
                cmd = ret.group(1)
                log.debug("matched command '%s' with expr type '%s'", cmd,
                          expr_type)
                return cmd

    @property
    @abc.abstractmethod
    def services(self):
        """ Return a dictionary of identified services and their state. """

    @property
    @abc.abstractmethod
    def processes(self):
        """
        Return a dictionary of processes associated with identified
        services.
        """

    @property
    @abc.abstractmethod
    def summary(self):
        """ Return a dictionary summary of this class i.e. services,
        their state and associated processes.
        """



from .common import (  # noqa: F403,F401
    OpenstackBase,
    OpenstackChecksBase,
    OpenstackEventHandlerBase,
    OpenstackEventCallbackBase,
)
from .openstack import OpenstackConfig  # noqa: F403,F401
